import requests
import json
import pandas as pd
from loguru import logger

from .utils.ut import Utils

class Authorization:

    def __init__(self, core:object) -> None:
        global client
        client = core

        self.raiseException = client.raiseException
        self.defaults = core.defaults
        self.structure = core.structure
        self.scheme = core.scheme

        self.header = {
            'Authorization': f'Bearer {client.accessToken}',
            'AcceptEncoding': 'deflate',
            'Accept': 'application/json',
            'Content-type': 'application/json',
            }

        self.baseUrl = f'https://{client.host}/authz'

    def roles(self) -> pd.DataFrame:
        """
        Returns a DataFrame of available roles
        """

        url = f'{self.baseUrl}/roles'
        response = requests.get(url=url, headers=self.header)
        return AuthUtils.handleResponse(self, response)

    def rules(self) -> pd.DataFrame:
        """
        Returns a DataFrame of available rules
        """

        url = f'{self.baseUrl}/rules'
        response = requests.get(url=url, headers=self.header)
        return AuthUtils.handleResponse(self, response)

    def users(self) -> pd.DataFrame:
        """
        Returns a DataFrame of available users
        """

        url = f'{self.baseUrl}/users'
        response = requests.get(url=url, headers=self.header)
        return AuthUtils.handleResponse(self, response)

    def userGroups(self) -> pd.DataFrame:
        """
        Returns a DataFrame of available users
        """

        url = f'{self.baseUrl}/usergroups'
        response = requests.get(url=url, headers=self.header)
        return AuthUtils.handleResponse(self, response)

    def createRole(
        self,
        inventoryName:str, 
        roleName:str,
        userGroups:list=None, 
        objectPermissions:list=['Create', 'Delete'], 
        propertiesPermissions:list=['Read', 'Update']
        ) -> None:

        """
        Creates a role and sets all rights to all propertes except references

        Parameters:
        ----------
        inventoryName : str
            The name of the inventory for which the new role authorizes rights.
        roleName : str
            Name of the new role.
        userGroup : list = None
            List of user group names. If None, the role will be created without attaching user groups.
        objectPermissions : list = ['Create', 'Delete']
            Default is 'Create' and 'Delete' to allow creating and deleting items of the specified inventory.
            Other entries are not allowed.
        propertiesPermissions : list = ['Read', 'Update']
            Default is 'Read' and 'Update'. All properties except references will receive 
            the here specified rights. Other entries are not allowed.     
        """

        # Parameter validation
        try:
            self.structure[inventoryName]
        except:
            Utils._error(self, f"Unknown inventory '{inventoryName}'")
            return

        roles = self.roles()
        if roleName in list(roles['name']):
            Utils._error(self, f"Role '{roleName}' already exists.")
            return

        if isinstance(userGroups, str):
            userGroups = [userGroups]

        if userGroups != None:
            dfUserGroups = self.userGroups()
            falseUserGroups = []
            for group in userGroups:
                if group not in list(dfUserGroups['name']):
                    falseUserGroups.append(group)
            
            if falseUserGroups:
                Utils._error(self, f"Unknown user group(s) {falseUserGroups}")
                return

        # Create role
        properties = client.structure[inventoryName]['properties']

        def createPropertyDict(properties):
            propertyList = []
            for _, value in properties.items():
                p = {}
                p.setdefault('id',value['propertyId'])
                p.setdefault('name',value['name'])
                p.setdefault('nullable',value['nullable'])
                p.setdefault('dataType',value['dataType'])
                p.setdefault('isArray',value['isArray'])
                propertyList.append(p)

            return propertyList

        def createPermissionsDict(properties):
            return {'scalarPropertiesPermissions': {key:propertiesPermissions for key in properties}}

        role = {
            'name': roleName,
            'objectPermissions': objectPermissions,
            'rootTypeName': inventoryName,
            'propertiesPermissions': createPermissionsDict(properties),
            'DynamicObjectTypeModel':
                [
                    {
                        'id':client.structure[inventoryName]['inventoryId'],
                        'name':inventoryName,
                        'isDomainUserType':client.structure[inventoryName]['isDomainUserType'],
                        'scalarProperties': createPropertyDict(properties),
                        'referencedProperties': []
                    }
                ]
        }
        
        urlRole = f'{self.baseUrl}/roles/'
        responseRole = requests.post(url=urlRole, headers=self.header, data=json.dumps(role))
        if responseRole.status_code not in [200, 204]:
            Utils._error(self, f"Status: {responseRole.status_code}, {json.loads(responseRole.content)}")
            return
        logger.info(f"Role {roleName} created.")

        # Create rules
        if userGroups != None:
            for group in userGroups:
                rule = {
                    'userGroup':group,
                    'roleRootType': inventoryName,
                    'role':roleName,
                }
                urlRule = f'{self.baseUrl}/rules/'
                responseRule = requests.post(url=urlRule, headers=self.header, data=json.dumps(rule))
                if responseRule.status_code not in [200, 204]:
                    Utils._error(self, f"Status: {responseRule.status_code}, {json.loads(responseRule.content)}")
                    return
                else:
                    logger.info(f"Rule for {roleName} and user group {group} created.")

        return

    def deleteRole(self, role:str) -> None:
        """
        Deletes a role and all related rules.
        """
        
        # Get Ids of roles and rules
        roles = self.roles().set_index('name')
        roleId = roles.loc[role,'id']

        rules = self.rules()
        rules = rules.set_index('role')
        try:
            ruleIds = rules.loc[role,'id']
        except:
            ruleIds = []
        if not isinstance(ruleIds, str):
            ruleIds = list(ruleIds)
        else:
            ruleIds = [ruleIds]

        # First delete rules
        if ruleIds:
            for ruleId in ruleIds:
                url = f'{self.baseUrl}/rules/{ruleId}'
                response = requests.delete(url=url, headers=self.header)
                if response.status_code not in [200, 204]:
                    Utils._error(self, f"Rule of role {roleId} could not be deleted. (Status {response.status_code})")
                    return 
                else:
                    logger.info(f"Rule of role {role} with id {ruleId} has been deleted.")     

        # After all rules have been deleted, delete the role
        url = f'{self.baseUrl}/roles/{roleId}'
        response = requests.delete(url=url, headers=self.header)
        if response.status_code not in [200, 204]:
            Utils._error(self, f"Role could not be deleted. (Status {response.status_code})")        
        else:
            logger.info(f"Role {role} with id {roleId} has been deleted.")   
            return


class AuthUtils:

    def handleResponse(client, response:object):
        if response.status_code not in [200, 204]:
            Utils._error(client, f"Status: {response.status_code}, {json.loads(response.content)}")        
        else:
            result = json.loads(response.content)
            return pd.json_normalize(result)


from loguru import logger
import requests
import os
from gql import gql, Client#, AIOHTTPTransport, RequestsHTTPTransport # This is gql version 3
from gql.transport.requests import RequestsHTTPTransport
import json

class FileImportService():
    
    def __init__(self) -> None:
        endpoint = 'http://172.16.2.206:7200/graphql/'
    
        transport =  RequestsHTTPTransport(url=endpoint, verify=True)
        self.client = Client(transport=transport, fetch_schema_from_transport=False)

    def importFile(self, path):
        file = open(path, 'rb')
        fileName = os.path.split(path)[-1]

        graphQlString = 'mutation ($input: [UploadFileInput!]!) { uploadFiles(input: $input) { ... on FileFailed { fileName message } ... on ProfileFailed { fileName importProfileId message } ... on ObjectFailed { fileName importProfileId messages } } }'

        files = {
            '0': (fileName, file),
            'operations': (None, f'{{ "query": "{graphQlString}", "variables": {{"input": [{{ "file": null }}]}}}}'),
            'map': (None, '{ "0": ["variables.input.0.file"] }'),
        }

        response = requests.post(self.endpoint, files=files)

        response = json.loads(response.text)
        for file in response['data']['uploadFiles']:
            logger.error(f"Filename: {file['fileName']}, ImportProfileId: {file['importProfileId']}, Message: {file['message']} ")

        return

if __name__ == '__main__':
    a = FileImportService()
    resp = a.importFile(r'C:\temp\techstack_test\items_appartments_comma.csv')


import pandas as pd
from datetime import datetime, timedelta
from gql import Client
from gql.transport.requests import RequestsHTTPTransport
from loguru import logger

from .utils.ut import Utils
from .utils.ut_autprog import AutProgUtils

class Schedule():
    def __init__(self, accessToken:str, endpoint:str, client:object) -> None:
              
        header = {
            'authorization': 'Bearer ' + Automation.accessToken
        }
        
        transport =  RequestsHTTPTransport(url=endpoint, headers=header, verify=True, proxies=client.proxies)
        self.client = Client(transport=transport, fetch_schema_from_transport=False)
        self.raiseException = client.raiseException
        self.defaults = client.defaults

    def schedules(
        self, 
        fields:list=None, 
        where:str=None) -> pd.DataFrame:
        """
        Returns schedules in a DataDrame

        Parameters:
        -----------
        fields: list | str = None
            A list of all properties to be queried. If None, all properties will be queried.
        where: str = None
            Use a string to add where criteria like
            ''workflowId eq "meteoData"'.

        Example:
        --------
        >>> Schedule.schedules(
                where='workflowId == "meteoData"', 
                fields=['name', 'cron']
            )   
        """

        key = 'schedules'

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''scheduleId
                name
                description
                workflowId
                businessKey
                cron
                isActive
                nextFireTime
                variables {{
                    key
                    value
                }}''' 

        resolvedFilter = ''
        if where != None: 
            resolvedFilter = f'({Utils._resolveWhereString(where)})'

        graphQLString = f'''query schedules {{
            {key}{resolvedFilter}  {{
                {_fields}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result[key])
        return df

    def createSchedule(self, name:str, workflowId:str, businessKey:str, cron:str, 
        isActive:bool=True, description:str=None, variables:dict=None) -> str:
        """Creates a schedule and returns the schedule Id"""

        if isActive == True:
            isActive = 'true'
        else:
            isActive = 'false'

        if description != None:
            description = description
        else: 
            description = ''

        if variables != None:
            _variables = 'variables: [\n'
            for k,v in variables.items():
                _variables += f'{{key: "{k}", value: "{v}"}}\n'
            _variables += ']'
        else: _variables = ''

        graphQLString = f'''mutation createSchedule {{
            createSchedule(input:{{
                name: "{name}"
                workflowId: "{workflowId}"
                businessKey: "{businessKey}"
                cron: "{cron}"
                description: "{description}"
                isActive: {isActive}
                {_variables}              
            }})
            {{
                schedule {{
                    scheduleId
                }}
                errors {{
                    message
                }}
            }}
        }}'''

        result = Utils._executeGraphQL(self, graphQLString)
        logger.debug(graphQLString)
        if result == None: return

        key = 'createSchedule'
        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
        else:
            scheduleId = result[key]['schedule']['scheduleId']
            logger.info(f"New schedule {scheduleId} created.")

        return scheduleId

    def updateSchedule(self, scheduleId, name:str=None, workflowId:str=None, businessKey:str=None, 
        cron:str=None, isActive:bool=None, description:str=None, variables:dict=None) -> None:
        """
        Updates a schedule. Only arguments that ar not None will overwrite respective fields.
        
        Parameters:
        -----------
        scheduleId : str
            The Id of the schedule that is to be updated.
        name : str
            The name of the schedule.
        workflowId : str
            The Id of the workflow that shall be executed with this schedule.
        cron : str
            The cron expression. For detailed information loop up
            http://www.quartz-scheduler.org/documentation/quartz-2.3.0/tutorials/crontrigger.html
        isActive : bool
            Determines, if the schedule should execute the workflow or not.
        description : str
            A description of the schedule.
        variables : dict
            A dictionary of variables that are used by tasks in the workflow.

        Example:
        --------
        >>> vars = {
                'var1': 99,
                'var2': "AnyString"
            }
        >>> client.Scheduler.updateSchedule('112880211090997248', name='test_schedule',
                isActive=True, variables=vars)
        
        """

        updateScheduleArgs = ''

        if name != None: updateScheduleArgs += f'name: "{name}"\n'
        if workflowId != None: updateScheduleArgs += f'workflowId: "{workflowId}"\n'
        if businessKey != None: updateScheduleArgs += f'businessKey: "{businessKey}"\n'
        if cron != None: updateScheduleArgs += f'cron: "{cron}"\n'
        if isActive != None: updateScheduleArgs += f'isActive: {str(isActive).lower()}\n'
        if description != None: updateScheduleArgs +=  f'description: "{description}"\n'

        if variables != None:
            _variables = 'variables: [\n'
            for k,v in variables.items():
                _variables += f'{{key: "{k}", value: "{v}"}}\n'
            _variables += ']'
            updateScheduleArgs += _variables

        graphQLString = f'''mutation updateSchedule {{
            updateSchedule(
                scheduleId: "{scheduleId}"
                input:{{
                    {updateScheduleArgs}
            }})
            {{
                errors {{
                    message
                }}
            }}
        }}'''

        result = Utils._executeGraphQL(self, graphQLString)
        logger.debug(graphQLString)
        if result == None: return

        key = 'updateSchedule'
        if result[key]['errors']:
            Utils._listGraphQlErrors(result, key)
        else:
            logger.info(f"Schedule {scheduleId} updated.")

        return

    def deleteSchedule(self, scheduleId:str, force:bool=False):
        """Deletes a schedule"""

        if force == False:
            confirm = input(f"Press 'y' to delete schedule '{scheduleId}': ")

        graphQLString = f'''mutation deleteSchedule {{
            deleteSchedule (scheduleId: "{scheduleId}")
            {{
                errors {{
                message
                }}
            }}
        }}
        '''      

        if force == True: confirm = 'y'
        if confirm == 'y':
            result = Utils._executeGraphQL(self, graphQLString)
            if result == None: return

            key = 'deleteSchedule'
            if result[key]['errors']:
                Utils._listGraphQlErrors(result, key)
            else:
                logger.info(f"Schedule {scheduleId} deleted")
                return None

    def nextFireTimes(self, workflowId:str, fromTimepoint:str=None, toTimepoint:str=None, count:int=None):
        """Show next fire times of a workflow"""

        if fromTimepoint == None:
            fromTimepoint = datetime.today().isoformat()

        if toTimepoint == None:
            toTimepoint = datetime.today() + timedelta(days=3)

        if count == None:
            _count = ''
        else: 
            _count = f'count: {count}'

        graphQLString = f'''query nextFireTimes {{
            nextFireTimes (
                workflowId: "{workflowId}",
                from: "{fromTimepoint}", 
                to: "{toTimepoint}",
                {_count}
                
                ) {{
                scheduleId
                fireTime
                }}
            }}
        '''
        
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result['nextFireTimes'])

        return df

class Automation():
    accessToken = ''

    def __init__(self, accessToken:str, endpoint:str, client:object) -> None:
        
        Automation.accessToken = accessToken
        
        header = {
            'authorization': 'Bearer ' + accessToken
        }
        
        transport =  RequestsHTTPTransport(url=endpoint, headers=header, verify=False, proxies=client.proxies)
        self.client = Client(transport=transport, fetch_schema_from_transport=True)
        self.raiseException = client.raiseException
        self.defaults = client.defaults

    def workflows(self) -> pd.DataFrame:
        """Returns a DataFrame of all Workflows"""

        graphQLString = f'''query workflows {{
            workflows {{
                id
                name
                description
                }}
            }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        df = pd.json_normalize(result['workflows'])
        return df

    def workflowInstances(self, workflowId:str=None, fromTimepoint:str=None, toTimepoint:str=None, 
        fields:list=None, where:str=None, showTasks=False) -> pd.DataFrame:
        """Shows Instances of a workflow. If workflowId=None, all Instances of all 
        workflows will be returned."""

        meta = ['id', 'name', 'businessKey', 'version', 'startTime', 'endTime', 'state']
        key = 'workflowInstances'

        if workflowId != None:
            _workflowId = f'workflowId: "{workflowId}"'
        else:
            _workflowId = ''

        if fromTimepoint != None:
            _fromTimepoint = f'from: "{fromTimepoint}"'
        else: 
            _fromTimepoint = ''

        if toTimepoint != None:
            _toTimepoint = f'from: "{toTimepoint}"'
        else: 
            _toTimepoint = ''

        if fields != None:
            if type(fields) != list:
                fields = [fields]
            _fields = Utils._queryFields(fields, recursive=True)   
        else:
            _fields =f'''
                id
                name
                businessKey
                version
                startTime
                endTime
                state
                variables {{
                    name
                    value
                    time
                }}''' 

        resolvedFilter = ''
        if where != None: 
            resolvedFilter = Utils._resolveWhereString(where)

        if showTasks != False:
            _tasks = f'''tasks {{
                            id
                            topic
                            workerId
                            timestamp
                            state
                            retries
                            errorMessage
                        }}'''
        else:
            _tasks = ''

        graphQLString = f'''query Instances {{
            {key}({_workflowId}, {_fromTimepoint}, {_toTimepoint}, {resolvedFilter}, all:true) {{
                {_fields}
                {_tasks}
                }}
            }}
            '''
     
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        if showTasks != False:
            df = pd.json_normalize(result[key], meta=meta, record_path=['tasks'], record_prefix='task.', errors='ignore')
            if 'startTime' in df.columns:
                df = df.sort_values(by='startTime', ascending=False)
        else:
            df = pd.json_normalize(result[key])
            if 'startTime' in df.columns:
                df = df.sort_values(by='startTime', ascending=False)
        return df
    
    def createWorkflow(self, id, name, description:str=None):

        graphQLString = f'''mutation createWorkflow {{
            createWorkflow(
                input: {{
                    id: "{id}"
                    name: "{name}"
                    description: "{description}"
                }}
                ) {{
                    ...on CreateWorkflowError {{
                    message
                    }}
                    ... on WorkflowCreated {{
                        workflow {{
                            id
                        }}
                    }}
                }}
            }}
        '''
        
        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        return result
                        
    def deployWorkflow(self, workflowId:str, filePath:str) -> None:
        """Deploys a Camunda XML to an existing workflow"""

        fileContent = Utils._encodeBase64(filePath)
        logger.debug(f"fileContent: {fileContent[:10]}")

        graphQLString = f'''mutation deployWorkflow {{
            deployWorkflow(
                input: {{
                    fileContentBase64: "{fileContent}"
                    workflowId: "{workflowId}"
                }}
            ) {{
                ... on DeployWorkflowError {{
                    message
                }}
                ... on InvalidWorkflowProcessId {{
                    processId
                    workflowId
                    message
                }}
                ... on WorkflowDeployed {{
                    version
                }}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        return result

    def startWorkflow(self, workflowId:str, businessKey:str, inputVariables:dict=None):
        """Starts a workflow"""

        if inputVariables == None:
            _vars = ''
        else:
            _vars = AutProgUtils._varsToString(inputVariables, 'input')

        graphQLString = f'''
            mutation ExecuteWF {{
                startWorkflow(input: {{ 
                    businessKey: "{businessKey}"
                    workflowId: "{workflowId}" 
                    {_vars}
                    }}
                ) {{
                    ... on ProcessDefinitionNotFound {{
                        workflowId
                        message
                        }}
                    ... on StartWorkflowError {{
                            message
                            }}
                    ... on WorkflowStarted {{
                        workflowInstanceId
                        }}
                    }}
                }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        if result == None: return

        return result

    def deleteWorkflow(self, workflowId:str):
        """Deletes a workflow"""

        graphQLString = f'''mutation deleteWorkflow {{
            deleteWorkflow (id: "{workflowId}")
            {{
                ... on DeleteWorkflowError {{
                    message
                    }}
                ...on WorkflowDeleted {{
                    success
                    }}
                ... on WorkflowNotFound {{
                    workflowId
                    message
                    }}
                
                }}
            }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        return result

    def terminateWorkflowInstance(self, workflowInstanceId):
        """Terminates a workflow instance"""

        graphQLString = f'''mutation terminateWorkflowInstance {{
            terminateWorkflowInstance(
                workflowInstanceId:"{workflowInstanceId}") {{
                ...on TerminateWorkflowInstanceError {{
                    message
                    }}
                ...on WorkflowInstanceNotFound {{
                    workflowInstanceId
                    message
                    }}
                ...on WorkflowInstanceTerminated {{
                    success
                    }}
                }}
            }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        return result

    def updateWorkflow(self, workflowId:str, name:str=None, description:str=None):
        """Updates a workflow (name and description can be changed)"""

        name = Utils._argNone('name', name)
        description = Utils._argNone('description', description)

        key = 'updateWorkflow'
        graphQLString = f'''mutation updateWorkflow {{
            {key}(workflowId: "{workflowId}", properties: {{
                {description}
                {name}
                }}) {{
                    ... on UpdateWorkflowError {{
                    message
                    }}
                    ... on WorkflowNotFound {{
                    workflowId
                    message
                    }}
                    ... on WorkflowUpdated {{
                    workflow {{
                        id
                        name
                        description
                        }}
                    }}
                }}
            }}
            '''

        result = Utils._executeGraphQL(self, graphQLString)
        return result

    def retryTask(self, externalTaskId):
        key = 'retryTask'

        graphQLString = f'''mutation retryTask {{
            {key}(externalTaskId: "{externalTaskId}") {{
                ... on RetryTaskError {{
                message
                }}
                ...on TaskNotFound {{
                message
                }}
                ... on TaskRetried {{
                success
                }}
            }}
        }}
        '''

        result = Utils._executeGraphQL(self, graphQLString)
        return result



from itertools import islice
import pandas as pd
from numpy import nan
import pytz
from loguru import logger
from .ut import Utils
import math

gq = Utils._toGraphQL

class UtilsTimeSeries():

    def _tsCollectionToString(collection:list) -> str:
        """Converts a list of dicts with time series data into graphQL string"""

        tsData = '[\n'
        for item in collection:
            tsData += '{\n'
            for key, value in item.items():
                if key in ['sys_inventoryId', 'sys_inventoryItemId']:
                    tsData += f'  {key}: "{value}"\n'
                elif key == 'data':
                    tsData += '  data: {\n'
                    for dkey, dvalue in value.items():
                        if dkey == 'resolution':
                            tsData += f'''  resolution: {{timeUnit: {dvalue['timeUnit']}, factor: {dvalue['factor']}}}\n'''
                        elif dkey == 'unit':
                            tsData += f'''  unit: "{dvalue}"\n'''
                        elif dkey == 'dataPoints':
                            tsData += '    dataPoints: [\n'
                            for dp_key in dvalue:
                                if 'flag' not in dp_key:
                                    value = dp_key['value']
                                    if (value is None) or (isinstance(value, float) and math.isnan(value)):
                                        tsData += f'''     {{timestamp: "{dp_key['timestamp']}", value: 0 flag: MISSING}}\n'''
                                    else:
                                        tsData += f'''     {{timestamp: "{dp_key['timestamp']}", value: {value}}}\n'''
                                else:    
                                    tsData += f'''     {{timestamp: "{dp_key['timestamp']}", value: {dp_key['value']}, flag: {dp_key['flag']}}}\n'''
                            tsData += '    ]\n'
                        else: pass
                    tsData += '  }\n'
                else:pass
            tsData += '}\n'
        tsData += ']\n'
        return tsData

    def _dataPointsToString(dataPoints:dict) -> str:
        """Converts a dictionary of datapoints to GraphQL string"""

        _dataPoints = ''
        for timestamp, value in dataPoints.items():
            if (value is None) or (isinstance(value, float) and math.isnan(value)):
                _dataPoints += f'{{ timestamp: {gq(timestamp)} value: 0 flag: MISSING }}\n'
            else: 
                _dataPoints += f'{{ timestamp: {gq(timestamp)} value: {gq(value)} }}\n'

        return _dataPoints

    def _sliceDataPoints(dataPoints:dict, start:int, stop:int):
        """Return a slice of the dataPoints dictionary"""

        return dict(islice(dataPoints, start, stop))

    def _processDataFrame(
        self,
        result:dict,
        dataFrame:pd.DataFrame,
        fields:list,
        timeZone:pytz.timezone,
        displayMode:str, 
        includeMissing:bool
        ) -> pd.DataFrame:
        """
        Creates a time Series DataFrame for different time series data methods
        """
        df = dataFrame
        if df.empty:
            logger.info('The query did not produce results.')
            return df
        df.loc[(df.flag == 'MISSING'), 'value'] = nan
        
        if displayMode == 'pivot':
            if includeMissing == False:
                df = df.pivot_table(index='timestamp', columns=fields, values='value', dropna=True)
            else:
                df = df.pivot_table(index='timestamp', columns=fields, values='value', dropna=False)

            if len(df.columns) == 0:
                logger.warning(f"Could not pivot DataFrame. Try different properties for fields-argument or use displayMode='rows'")
                return
            df.index = pd.to_datetime(df.index, format='%Y-%m-%dT%H:%M:%S').tz_convert(pytz.timezone(timeZone))
            if self.defaults.useDateTimeOffset == False:
                df.index = df.index.tz_localize(tz=None)

            columnNumber = len(result)
            dfColumnNumber = len(df.columns)
            if len(df.columns) < len(result):
                logger.info(f"{columnNumber-dfColumnNumber} columns omitted due to duplicate column headers or NaN-columns.")
            return df

        elif displayMode == 'rows':
            df.set_index('timestamp', inplace=True)
            df.index = pd.to_datetime(df.index, format='%Y-%m-%dT%H:%M:%S').tz_convert(pytz.timezone(timeZone))
            if self.defaults.useDateTimeOffset == False:
                df.index = df.index.tz_localize(tz=None)
            return df

        else:
            Utils._error(self, msg=f"Unknown displayMode '{displayMode}'. Use 'pivot' or 'rows'.")
            return

 